# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Notebooklet base classes."""
from pathlib import Path
from typing import Any, Dict, List, Optional, Set, Tuple, Union

import attr
import yaml
from attr import Factory

from ._version import VERSION

__version__ = VERSION
__author__ = "Ian Hellen"


@attr.s(auto_attribs=True)
class NBMetadata:
    """Notebooklet metadata class."""

    name: str = "Unnamed"
    mod_name: str = ""
    description: str = ""
    default_options: List[Union[str, Dict]] = Factory(list)
    other_options: List[Union[str, Dict]] = Factory(list)
    inputs: List[str] = ["value"]
    entity_types: List[str] = Factory(list)
    keywords: List[str] = Factory(list)
    req_providers: List[str] = Factory(list)

    # pylint: disable=not-an-iterable
    @property
    def search_terms(self) -> Set[str]:
        """Return set of search terms for the object."""
        return set(
            [self.name]
            + [obj.casefold() for obj in self.entity_types]  # type: ignore
            + [key.casefold() for key in self.keywords]  # type: ignore
            + [opt.casefold() for opt in self.all_options]  # type: ignore
        )

    def __str__(self):
        """Return string representation of object."""
        return "\n".join(f"{name}: {val}" for name, val in attr.asdict(self).items())

    @property
    def all_options(self) -> List[str]:
        """Return combination of default and other options."""
        opts = []
        if self.default_options:
            for opt in self.default_options:
                if isinstance(opt, str):
                    opts.append(opt)
                elif isinstance(opt, dict):
                    opts.append(next(iter(opt.keys())))
        if self.other_options:
            for opt in self.other_options:
                if isinstance(opt, str):
                    opts.append(opt)
                elif isinstance(opt, dict):
                    opts.append(next(iter(opt.keys())))
        return sorted(opts)

    def get_options(self, option_set: str = "all") -> List[Tuple[str, Optional[str]]]:
        """
        Return list of options and descriptions.

        Parameters
        ----------
        option_set : str, optional
            The subset of options to return, by default "all"
            Other values are "default" and "other"

        Returns
        -------
        List[Tuple[str, Optional[str]]]
            A list of tuples of option name and description.

        """
        opt_list: List[Tuple[str, Optional[str]]] = []
        if option_set.casefold() in ["all", "default"] and self.default_options:
            for opt in self.default_options:
                if isinstance(opt, str):
                    opt_list.append((opt, None))
                elif isinstance(opt, dict):
                    opt_list.extend(opt.items())
        if option_set.casefold() in ["all", "other"] and self.other_options:
            for opt in self.other_options:
                if isinstance(opt, str):
                    opt_list.append((opt, None))
                elif isinstance(opt, dict):
                    opt_list.extend(opt.items())
        return opt_list

    @property
    def options_doc(self) -> str:
        """Return list of options and documentation."""
        def_options = self.get_options("default")

        opt_list = [
            "",
            "    Default Options",
            "    ---------------",
        ]
        if def_options:
            opt_list.extend([f"    - {key}: {value}" for key, value in def_options])
        else:
            opt_list.append("    None")
        opt_list.extend(
            [
                "",
                "    Other Options",
                "    -------------",
            ]
        )

        if self.get_options("other"):
            opt_list.extend(
                [f"    - {key}: {value}" for key, value in self.get_options("other")]
            )
        else:
            opt_list.append("    None")
        # Add a blank line to the end
        opt_list.extend(["", ""])
        return "\n".join(opt_list)

    # pylint: enable=not-an-iterable


def read_mod_metadata(mod_path: str, module_name) -> Tuple[NBMetadata, Dict[str, Any]]:
    """
    Read notebooklet metadata from yaml file.

    Parameters
    ----------
    mod_path : str
        The fully-qualified (dotted) module name
    module_name : str
        The full module name.

    Returns
    -------
    Tuple[NBMetadata, Dict[str, Any]]
        A tuple of the metadata class
        and the documentation dictionary

    """
    md_dict = _read_metadata_file(mod_path)
    if not md_dict:
        return NBMetadata(), {}
    metadata_vals = md_dict.get("metadata", {})

    metadata_vals["mod_name"] = module_name
    metadata = NBMetadata(**metadata_vals)
    output = md_dict.get("output", {})
    return metadata, output


def _read_metadata_file(mod_path):
    md_path = Path(str(mod_path).replace(".py", ".yaml"))
    if not md_path.is_file():
        md_path = Path(str(mod_path).replace(".py", ".yml"))
    if md_path.is_file():
        with open(md_path, "r", encoding="utf-8") as _md_file:
            return yaml.safe_load(_md_file)
    return None


def update_class_doc(cls_doc: str, cls_metadata: NBMetadata):
    """Append the options documentation to the `cls_doc`."""
    options_doc = cls_metadata.options_doc
    if options_doc is not None:
        return cls_doc + options_doc
    return cls_doc



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Data Providers class and init function."""
import inspect
import logging
import sys
from collections import namedtuple
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union

from msticpy.common.exceptions import MsticpyAzureConfigError
from msticpy.common.wsconfig import WorkspaceConfig

try:
    from msticpy.context import GeoLiteLookup, IPStackLookup, TILookup
    from msticpy.context.azure.azure_data import AzureData
    from msticpy.data.core.data_providers import QueryProvider
    from msticpy.data.core.query_defns import DataEnvironment
except ImportError:
    # Fall back to msticpy locations prior to v2.0.0
    from msticpy.data.azure.azure_data import AzureData
    from msticpy.data import QueryProvider
    from msticpy.data.query_defns import DataEnvironment
    from msticpy.sectools import GeoLiteLookup, IPStackLookup, TILookup

from ._version import VERSION
from .common import MsticnbDataProviderError, MsticnbError
from .options import get_opt

__version__ = VERSION
__author__ = "Ian Hellen"


class SingletonDecorator:
    """
    Singleton decorator class.

    Notes
    -----
    Using this decorator on a class enforces the following
    behavior:

    - First instantiation of class will work as normal
    - Subsequent attempts with the same set/values of kwargs
      will just return the original class
    - Instantiation of the class with a different set of kwargs
      will instantiate a new class.
    - The class method `current()` will always return the
      last instance of the class.

    """

    def __init__(self, wrapped_cls):
        """Instantiate the class wrapper."""
        self.wrapped_cls = wrapped_cls
        self.instance = None
        self.__doc__ = wrapped_cls.__doc__

    def __call__(self, *args, **kwargs):
        """Overide the __call__ method for the wrapper class."""
        if (
            self.instance is None
            or getattr(self.instance, "kwargs", None) != kwargs
            or getattr(self.instance, "args", None) != args
        ):
            self.instance = self.wrapped_cls(*args, **kwargs)
            self.instance.kwargs = kwargs
            self.instance.args = args
        return self.instance

    def current(self):
        """Return the current instance of the wrapped class."""
        return self.instance

    def __getattr__(self, name):
        """Return the attribute `name` from the wrapped class."""
        if hasattr(self.wrapped_cls, name):
            return getattr(self.wrapped_cls, name)
        if self.instance is None:
            return None
        if name == "current":
            return self.instance
        return getattr(self.instance, name)


ProviderDefn = namedtuple("ProviderDefn", "prov_class, connect_reqd, get_config")


@SingletonDecorator
class DataProviders:
    """Notebooklet DataProviders class."""

    _DEFAULT_PROVIDERS = ["tilookup", "geolitelookup"]
    _OTHER_PROVIDERS = ["ipstacklookup"]

    def __init__(
        self,
        query_provider: Union[str, QueryProvider] = "MSSentinel",
        providers: Optional[List[str]] = None,
        **kwargs,
    ):
        """
        Instantiate an singleton instance of DataProviders.

        Parameters
        ----------
        query_provider : Union[str, QueryProvider], optional
            DataEnvironment name of the primary query provider,
            or an instance of an existing query provider,
            by default "MSSentinel"
        providers : Optional[List[str]], optional
            A list of provider names to load.
            You can add additional query providers by including them
            in the `providers` list.

        Other Parameters
        ----------------
        kwargs
            Pass parameters to individual providers using the following notation:
            `{provider_name}_{param_name}="param_value"
            Where `provider_name` is the name of the data provider,
            `param_name` is the parameter name expected by the
            provider and `param_value` is the value to assign to
            `param_name`. `param_value` can be any type.

            Depending on the provider, these parameters (with the prefix stripped)
            are sent to either the constructor or the `connect` method.

        Notes
        -----
        To see a list of current providers use `get_providers` class
        method.

        """
        self.provider_names: set = self._get_custom_providers(providers)
        self.provider_classes: Dict[str, ProviderDefn] = self._create_provider_defns()
        self.providers: Dict[str, Any] = {}

        self.query_provider = None
        parsed_provider = DataEnvironment.Unknown
        if isinstance(query_provider, str):
            parsed_provider = self._parse_provider_name(query_provider)
            self.provider_names.add(parsed_provider.name)
        elif isinstance(query_provider, QueryProvider):
            # If this is a query provider instance, just add it directly
            self.query_provider = query_provider
            parsed_provider = query_provider.environment
            self.providers[parsed_provider] = query_provider

        # Go through list of providers, instantiating and connecting them.
        for provider in sorted(self.provider_names):
            try:
                self.add_provider(provider, **kwargs)
            except MsticnbDataProviderError as err:
                print(f"Data provider {provider} could not be added.")
                print(err.args)
            else:
                if (
                    provider in self.providers
                    and not self.query_provider
                    and provider == parsed_provider.name
                ):
                    # This is the default query provider
                    setattr(self, "query_provider", self.providers[provider])

    def __getitem__(self, key: str):
        """Return provider matching `key`."""
        if key in self.providers:
            return self.providers[key]
        alt_key = DataEnvironment.parse(key)
        if alt_key in self.providers:
            return self.providers[alt_key]
        raise KeyError(key, "not found")

    def _create_provider_defns(self):
        """Definitions for provider construction."""
        defns = {
            "azuresentinel": ProviderDefn(QueryProvider, True, self._azsent_get_config),
            "queryprovider": ProviderDefn(QueryProvider, True, None),
            "azuredata": ProviderDefn(AzureData, True, None),
            "tilookup": ProviderDefn(TILookup, False, None),
            "geolitelookup": ProviderDefn(GeoLiteLookup, False, None),
            "ipstacklookup": ProviderDefn(IPStackLookup, False, None),
        }
        # Add loganalytics as an alias for azuresentinel
        defns["loganalytics"] = defns["azuresentinel"]
        return defns

    @staticmethod
    def _parse_provider_name(query_provider):
        parsed_provider = DataEnvironment.parse(query_provider)
        # If we weren't able to match the name to a known provider, raise exception.
        if parsed_provider == DataEnvironment.Unknown:
            known_providers = set(DataEnvironment.__members__.keys()) - {
                "Unknown",
                "Kusto",
                "AzureSecurityCenter",
            }
            raise MsticnbDataProviderError(
                f"Unknown query provider '{query_provider}",
                f"Available providers are {', '.join(known_providers)}",
            )
        return parsed_provider

    def _get_custom_providers(self, providers):
        requested_provs = set(self._DEFAULT_PROVIDERS)
        if not providers:
            return requested_provs

        providers = [prov.casefold() for prov in providers]
        add_provs = {opt[1:] for opt in providers if opt.startswith("+")}
        sub_provs = {opt[1:] for opt in providers if opt.startswith("-")}
        std_provs = {opt for opt in providers if opt[0] not in ("+", "-")}
        if std_provs and (add_provs or sub_provs):
            raise MsticnbError(
                "Provider list must be either a list of providers to use",
                "or providers to add/remove from the default set.",
                "You cannot mix these.",
            )
        if sub_provs:
            requested_provs = requested_provs - sub_provs
        if add_provs:
            requested_provs = requested_provs | add_provs
        if not add_provs and not sub_provs:
            requested_provs = set(providers)

        return requested_provs

    def add_provider(self, provider: str, **kwargs):
        """
        Add a provider to the current provider set.

        Parameters
        ----------
        provider : str
            Name of the provider.

        """
        if provider in self.providers:
            return
        provider_key = provider.casefold()
        new_provider = None
        if provider in DataEnvironment.__members__:
            # If this is a known query provider, pass to appropriate
            prov_def = self.provider_classes.get(
                provider_key, self.provider_classes["queryprovider"]
            )
            new_provider = self._query_prov(provider, prov_def, **kwargs)
        elif provider_key in self.provider_classes:
            # Else if this is a known data provider
            prov_def = self.provider_classes[provider_key]
            if prov_def.connect_reqd:
                new_provider = self._query_prov(provider, prov_def, **kwargs)
            else:
                new_provider = self._no_connect_prov(provider, prov_def, **kwargs)
        else:
            raise MsticnbDataProviderError(f"Provider {provider} not recognized.")

        if new_provider:
            setattr(self, provider, new_provider)
            self.providers[provider] = new_provider

    def has_required_providers(
        self, reqd_provs: Iterable[str]
    ) -> Tuple[List[str], List[str]]:
        """
        Check the provider list against the list of loaded providers.

        Parameters
        ----------
        reqd_provs : Iterable[str]
            Required provider list.

        Returns
        -------
        Tuple[List[str], List[str]
            List of missing providers found
            List of unknown providers found

        """
        if not reqd_provs:
            return [], []
        r_prov_names = {prov for prov in reqd_provs if "|" not in prov}

        # remove matching items
        missing_provs = r_prov_names - set(self.providers)
        matched_provs = set()
        for req_prov in missing_provs:
            # See if the required provider is an alias of another query provider
            alt_name = DataEnvironment.parse(req_prov)
            if alt_name.name in missing_provs:
                matched_provs.add(req_prov)
                continue
            # If the required provider is a query provider
            # and the LocalData provider is loaded - assume this is
            # an intended replacement
            if alt_name != DataEnvironment.Unknown and "LocalData" in self.providers:
                matched_provs.add(req_prov)
        missing_provs = missing_provs - matched_provs

        # Process required providers where optional items are specified
        for r_prov in {prov for prov in reqd_provs if "|" in prov}:
            prov_opts = {prov.strip() for prov in r_prov.split("|")}
            # Add canonical name for any aliases
            alias_opts = {DataEnvironment.parse(prov).name for prov in prov_opts}
            prov_opts.update(alias_opts - {"Unknown"})
            # We only need to match one of these
            if any(m_prov for m_prov in prov_opts if m_prov in self.providers):
                continue
            missing_provs.add(r_prov)
        unknown_provs = missing_provs - set(self.list_providers())

        return list(missing_provs), list(unknown_provs)

    @classmethod
    def list_providers(cls) -> List[str]:
        """
        Return the list of all providers.

        Returns
        -------
        List[str]
            List of all providers.

        """
        providers = list(DataEnvironment.__members__.keys())
        providers.remove("Unknown")
        providers.extend(cls._DEFAULT_PROVIDERS)
        providers.extend(cls._OTHER_PROVIDERS)
        return providers

    # Provider initializers

    @classmethod
    def get_def_providers(cls) -> List[str]:
        """
        Return the list of default providers.

        Returns
        -------
        List[str]
            List of default providers.

        """
        return cls._DEFAULT_PROVIDERS

    def _query_prov(self, provider: str, provider_defn: ProviderDefn, **kwargs) -> Any:
        try:
            # Get any keys with the provider prefix and initialize the provider
            prov_kwargs_args = self._get_provider_kwargs(provider, **kwargs)

            # instantiate the provider class (sending all kwargs)
            provider_class = provider_defn.prov_class
            if (
                "data_environment"
                in inspect.signature(provider_class.__init__).parameters
            ):
                prov_kwargs_args = {"data_environment": provider, **prov_kwargs_args}
            logging.info(
                "Creating provider %s with args %s", provider, prov_kwargs_args
            )
            created_provider = provider_defn.prov_class(**prov_kwargs_args)
            if created_provider.connected:
                return created_provider
            # get the args required by connect function
            prov_connect_args = self._get_connect_args(
                created_provider.connect, **prov_kwargs_args
            )
            # if there is a function to get config settings
            if not prov_connect_args and provider_defn.get_config:
                prov_connect_args = provider_defn.get_config()
            # call the connect function
            try:
                created_provider.connect(**prov_connect_args)
            except Exception as err:  # pylint: disable=broad-except
                print(f"Connection attempt for {provider} failed.\n{err}")
            return created_provider
        except MsticpyAzureConfigError as mp_ex:
            if get_opt("verbose"):
                print("Warning:", mp_ex.args)
            return None

    def _no_connect_prov(
        self, provider: str, provider_defn: ProviderDefn, **kwargs
    ) -> Any:
        # Get the args passed to __init__ for this provider
        prov_args = self._get_provider_kwargs(provider, **kwargs)
        # If there are none and there's a config function, call that.
        if not prov_args and provider_defn.get_config:
            prov_args = provider_defn.get_config()
        # Instantiate the provider
        return provider_defn.prov_class(**prov_args)

    # Helper methods
    @staticmethod
    def _get_provider_kwargs(prefix: str, **kwargs) -> Dict[str, str]:
        """Return the kwargs prefixed with "prefix_"."""
        if prefix == "LogAnalytics" and any(
            key for key in kwargs if key.startswith("AzureSentinel")
        ):
            azsent_args = {
                key.replace("AzureSentinel", "LogAnalytics"): val
                for key, val in kwargs.items()
            }
            kwargs.update(azsent_args)

        return {
            name.replace(f"{prefix}_", ""): arg
            for name, arg in kwargs.items()
            if name.startswith(prefix)
        }

    @staticmethod
    def _get_connect_args(func: Callable, **kwargs) -> Dict[str, str]:
        """Get the arguments required by the `connect` function."""
        connect_params = inspect.signature(func).parameters
        return {
            name: arg_val for name, arg_val in kwargs.items() if name in connect_params
        }

    # Provider get_config functions
    @staticmethod
    def _azsent_get_config(**kwargs) -> Dict[str, str]:
        if "workspace" in kwargs:
            ws_config = WorkspaceConfig(workspace=kwargs["workspace"])
        elif "config_file" in kwargs:
            ws_config = WorkspaceConfig(config_file=kwargs["config_file"])
        elif (
            WorkspaceConfig.CONF_TENANT_ID_KEY in kwargs
            and WorkspaceConfig.CONF_WS_ID_KEY in kwargs
        ):
            ws_config = WorkspaceConfig()
            ws_config[WorkspaceConfig.CONF_TENANT_ID_KEY] = kwargs[
                WorkspaceConfig.CONF_TENANT_ID_KEY
            ]
            ws_config[WorkspaceConfig.CONF_WS_ID_KEY] = kwargs[
                WorkspaceConfig.CONF_WS_ID_KEY
            ]
        else:
            ws_config = WorkspaceConfig()

        if not ws_config.config_loaded:
            raise MsticnbDataProviderError(
                "Could not find valid MS Sentinel configuration.",
                "Please ensure configuration files are set correctly or supply",
                "azure_sentinel.workspace_id and azure_sentinel.tenant_id",
                "arguments to this class.",
            )
        return {"connection_str": ws_config.code_connect_str}


def init(
    query_provider: Union[str, QueryProvider] = "MSSentinel",
    providers: Optional[List[str]] = None,
    **kwargs,
):
    """
    Instantiate an instance of DataProviders.

    Parameters
    ----------
    query_provider : Union[str, QueryProvider], optional
        DataEnvironment name of the primary query provider,
        or an instance of an existing query provider,
        by default "MSSentinel"
        You can add additional query providers by including them
        in the `providers` list.
    providers : Optional[List[str]], optional
        A list of provider names, by default None

    Other Parameters
    ----------------
    kwargs
        You can pass parameters to individual providers using
        the following notation:
        `ProviderName_param_name="param_value"
        Where `ProviderName` is the name of the data provider,
        `param_name` is the parameter name expected by the
        provider and `param_value` is the value to assign to
        `param_name`. `param_value` can be any type.

        Depending on the provider, these parameters (with the
        prefix stripped) are sent to either the constructor or
        `connect` method.

    Notes
    -----
    To see a list of currently supported providers call:
    `DataProviders.list_providers()`

    """
    d_provs = DataProviders(query_provider, providers, **kwargs)
    print(f"Notebooklets: Loaded providers: {', '.join(d_provs.providers.keys())}")
    msticnb = sys.modules["msticnb"]
    setattr(msticnb, "data_providers", d_provs.providers)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
Template notebooklet.

Notebooklet modules have three main sections:

- **Result class definition**:
  This defines the attributes and descriptions of the data that you
  want to return from the notebooklet.
- **Notebooklet class definition**:
  This is the entry point for running the notebooklet. At minimum
  it should be a class derived from Notebooklet that implements
  a `run` method and returns your result class.
- **Functions**:
  These do most of the work of the notebooklet and usually the code
  that is copied from or adapted from the original notebook.

Having the latter section is optional. You can choose to implement
this functionality in instance methods of the notebooklet class.

However, there are advantages to keeping these as separate functions
outside the class. It means that all the data used in the functions
has to be passed around as parameters and return values. This can
improve the clarity of the code and reduce errors due to some
dependency on some mysterious global state.

If the user of your notebooklet wants to import the module's code
into a notebook to read and possibly adapt it, having standalone
functions will make it easier from them understand and work with
the code.

"""
# pylint: disable=duplicate-code
from typing import Any, Dict, Iterable, Optional, Union

import pandas as pd
from bokeh.models import LayoutDOM
from msticpy.common.timespan import TimeSpan

try:
    from msticpy.vis.timeline import display_timeline
except ImportError:
    # Fall back to msticpy locations prior to v2.0.0
    from msticpy.nbtools.nbdisplay import display_timeline

from ... import nb_metadata

# change the ".." to "...."
from ..._version import VERSION

# Note - when moved to the final location (e.g.
# nb/environ/category/mynotebooklet.py)
# you will need to change the "..." to "...." in these
# imports because the relative path has changed.
from ...common import (
    MsticnbMissingParameterError,
    nb_data_wait,
    nb_markdown,
    nb_print,
    set_text,
)

# change the "..." to "...."
from ...notebooklet import NBMetadata, Notebooklet, NotebookletResult

__version__ = VERSION
__author__ = "Your name"


# Read module metadata from YAML
_CLS_METADATA: NBMetadata
_CELL_DOCS: Dict[str, Any]
_CLS_METADATA, _CELL_DOCS = nb_metadata.read_mod_metadata(__file__, __name__)


# pylint: disable=too-few-public-methods
# Rename this class
class TemplateResult(NotebookletResult):
    """
    Template Results.

    Attributes
    ----------
    all_events : pd.DataFrame
        DataFrame of all raw events retrieved.
    plot : bokeh.models.LayoutDOM
        Bokeh plot figure showing the account events on an
        interactive timeline.
    additional_info: dict
        Additional information for my notebooklet.

    """

    def __init__(
        self,
        description: Optional[str] = None,
        timespan: Optional[TimeSpan] = None,
        notebooklet: Optional["Notebooklet"] = None,
    ):
        """
        Create new Notebooklet result instance.

        Parameters
        ----------
        description : Optional[str], optional
            Result description, by default None
        timespan : Optional[TimeSpan], optional
            TimeSpan for the results, by default None
        notebooklet : Optional[, optional
            Originating notebooklet, by default None

        """
        super().__init__(description, timespan, notebooklet)
        self.description: str = "Windows Host Security Events"

        # Add attributes as needed here.
        # Make sure they are documented in the Attributes section
        # above.
        self.all_events: Optional[pd.DataFrame] = None
        self.plot: Optional[LayoutDOM] = None
        self.additional_info: Optional[dict] = None


# pylint: enable=too-few-public-methods


# Rename this class
class TemplateNB(Notebooklet):
    """
    Template Notebooklet class.

    Detailed description of things this notebooklet does:

    - Fetches all events from XYZ
    - Plots interesting stuff
    - Returns extended metadata about the thing

    Document the options that the Notebooklet takes, if any,
    Use these control which parts of the notebooklet get run.

    """

    # assign metadata from YAML to class variable
    metadata = _CLS_METADATA
    __doc__ = nb_metadata.update_class_doc(__doc__, metadata)
    _cell_docs = _CELL_DOCS

    # @set_text decorator will display the title and text every time
    # this method is run.
    # The key value refers to an entry in the `output` section of
    # the notebooklet yaml file.
    @set_text(docs=_CELL_DOCS, key="run")
    def run(
        self,
        value: Any = None,
        data: Optional[pd.DataFrame] = None,
        timespan: Optional[TimeSpan] = None,
        options: Optional[Iterable[str]] = None,
        **kwargs,
    ) -> TemplateResult:
        """
        Return XYZ summary.

        Parameters
        ----------
        value : str
            Host name - The key for searches - e.g. host, account, IPaddress
        data : Optional[pd.DataFrame], optional
            Alternatively use a DataFrame as input.
        timespan : TimeSpan
            Timespan for queries
        options : Optional[Iterable[str]], optional
            List of options to use, by default None.
            A value of None means use default options.
            Options prefixed with "+" will be added to the default options.
            To see the list of available options type `help(cls)` where
            "cls" is the notebooklet class or an instance of this class.

        Returns
        -------
        TemplateResult
            Result object with attributes for each result type.

        Raises
        ------
        MsticnbMissingParameterError
            If required parameters are missing

        """
        # This line use logic in the superclass to populate options
        # (including default options) into this class.
        super().run(
            value=value, data=data, timespan=timespan, options=options, **kwargs
        )

        if not value:
            raise MsticnbMissingParameterError("value")
        if not timespan:
            raise MsticnbMissingParameterError("timespan.")

        # Create a result class
        result = TemplateResult(
            notebooklet=self, description=self.metadata.description, timespan=timespan
        )

        # You might want to always do some tasks irrespective of
        # options sent
        all_events_df = _get_all_events(
            self.query_provider, host_name=value, timespan=timespan
        )
        result.all_events = all_events_df

        if "plot_events" in self.options:
            result.plot = _display_event_timeline(event_data=all_events_df)

        if "get_metadata" in self.options:
            result.additional_info = _get_metadata(host_name=value, timespan=timespan)

        # Assign the result to the _last_result attribute
        # so that you can get to it without having to re-run the operation
        self._last_result = result  # pylint: disable=attribute-defined-outside-init

        return self._last_result

    # You can add further methods to do things after (or before) the main
    # run method. You might need these if you want to add an interaction
    # point where the user needs to select and option. For example, you
    # could have a "select_account" method that uses a widget to let the
    # notebook user pick from a list. Then have a follow on method that
    # does something with this choice.
    def run_additional_operation(
        self, event_ids: Optional[Union[int, Iterable[int]]] = None
    ) -> Optional[pd.DataFrame]:
        """
        Addition method.

        Parameters
        ----------
        event_ids : Optional[Union[int, Iterable[int]]], optional
            Single or interable of event IDs (ints).

        Returns
        -------
        pd.DataFrame
            Results with expanded columns.

        """
        # Include this to check the "run()" has happened before this method
        # can be run
        if (
            not self._last_result or self._last_result.all_events is None
        ):  # type: ignore
            print(
                "Please use 'run()' to fetch the data before using this method.",
                "\nThen call 'expand_events()'",
            )
            return None
        # Print a status message - this will not be displayed if
        # the user has set the global "verbose" option to False.
        nb_print("We maybe about to wait some time")

        nb_markdown("Print some message that always displays", "blue, bold")
        return _do_additional_thing(
            evt_df=self._last_result.all_events,  # type: ignore
            event_ids=event_ids,
        )
        # Note you can also assign new items to the result class in
        # self._last_result and return the updated result class.


# This section contains functions that do the work. It can be split into
# cells recognized by some editors (like VSCode) but this is optional


# %%
# Get Windows Security Events
def _get_all_events(qry_prov, host_name, timespan):
    # Tell the user that you're fetching data
    # (doesn't display if nb.set_opt("silent", True))
    nb_data_wait("SecurityEvent")

    # example real query
    # return qry_prov.WindowsSecurity.list_host_events(
    #     timespan,
    #     host_name=host_name,
    #     add_query_items="| where EventID != 4688 and EventID != 4624",
    # )

    # dummy query
    del qry_prov, host_name, timespan
    return pd.DataFrame(
        {
            "TimeGenerated": pd.date_range("2022-01-01", periods=5, tz="utc", freq="h"),
            "EventID": [4688, 4688, 4625, 4624, 4624],
            "Computer": ["MyHost.dom"] * 5,
            "Account": [f"user{n}" for n in range(5)],
            "Activity": [f"activity_{s}" for s in "abcde"],
        }
    )


# You can add title and/or text to individual functions as they run.
# You can reference text from sections in your YAML file or specify
# it inline (see later example)
@set_text(docs=_CELL_DOCS, key="display_event_timeline")
def _display_event_timeline(event_data):
    # Plot events on a timeline

    return display_timeline(
        data=event_data,
        group_by="EventID",
        source_columns=["Activity", "Account"],
        legend="right",
    )


# This function has no text output associated with it
def _get_metadata(host_name, timespan):
    return {
        "host": host_name,
        "data_items": {"age": 97, "color": "blue", "country_of_origin": "Norway"},
        "provider": "whois",
        "time_duration": timespan,
    }


# %%
# Extract event details from events
# Note using inline text output here - usually better to store this
# all in the yaml file for maintainability.
@set_text(
    title="Do something else",
    hd_level=3,
    text="""
This may take some time to complete for large numbers of events.

It will do:
- Item one
- Item two
""",
    md=True,
)
def _do_additional_thing(evt_df, event_ids):
    # nb_print is the same as print() except it honors the
    # 'silent' option.
    nb_print("Doing something time-consuming...")
    return evt_df[evt_df["EventID"].isin(event_ids)]



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""logons_summary - provides overview of host logon events."""
import re
from math import pi
from typing import Any, Dict, Iterable, Optional, Union

import pandas as pd
from bokeh.io import output_notebook
from bokeh.palettes import viridis  # pylint: disable=no-name-in-module
from bokeh.plotting import figure, show
from bokeh.transform import cumsum
from IPython.display import display
from msticpy.common.timespan import TimeSpan
from msticpy.common.utility import md

try:
    from msticpy.vis import timeline
    from msticpy.vis.foliummap import FoliumMap
except ImportError:
    # Fall back to msticpy locations prior to v2.0.0
    from msticpy.nbtools import timeline
    from msticpy.nbtools.foliummap import FoliumMap

from ...._version import VERSION
from ....common import MsticnbMissingParameterError, nb_data_wait, nb_print, set_text
from ....nb_metadata import read_mod_metadata
from ....nblib.azsent.host import verify_host_name
from ....notebooklet import NBMetadata, Notebooklet, NotebookletResult

pd.options.mode.chained_assignment = None

__version__ = VERSION
__author__ = "Pete Bryan"


_CLS_METADATA: NBMetadata
_CELL_DOCS: Dict[str, Any]
_CLS_METADATA, _CELL_DOCS = read_mod_metadata(__file__, __name__)


# pylint: disable=too-few-public-methods
class HostLogonsSummaryResult(NotebookletResult):
    """
    Host Logons Summary Results.

    Attributes
    ----------
    logon_sessions: pd.DataFrame
        A Dataframe summarizing all sucessfull and failed logon attempts
        observed during the specified time period.

    logon_map: FoliumMap
        A map showing remote logon attempt source locations. Red points
        represent failed logons, green successful.

    plots: Dict
        A collection of Bokeh plot figures showing various aspects of observed logons.
        Keys are a descriptive name of the plot and values are the plot figures.

    """

    def __init__(
        self,
        description: Optional[str] = None,
        timespan: Optional[TimeSpan] = None,
        notebooklet: Optional["Notebooklet"] = None,
    ):
        """
        Create new Notebooklet result instance.

        Parameters
        ----------
        description : Optional[str], optional
            Result description, by default None
        timespan : Optional[TimeSpan], optional
            TimeSpan for the results, by default None
        notebooklet : Optional[, optional
            Originating notebooklet, by default None

        """
        super().__init__(description, timespan, notebooklet)
        self.logon_sessions: Optional[pd.DataFrame] = None
        self.logon_matrix: Optional[pd.DataFrame] = None
        self.logon_map: Optional[FoliumMap] = None
        self.timeline: Optional[figure] = None
        self.failed_success: Optional[pd.DataFrame] = None
        self.plots: Optional[Dict[str, figure]] = None


class HostLogonsSummary(Notebooklet):  # pylint: disable=too-few-public-methods
    """
    Host Logons Summary Notebooket class.

    Queries and displays information about logons to a host including:

    - Summary of sucessfull logons
    - Visualizations of logon event times
    - Geolocation of remote logon sources
    - Visualizations of various logon elements depending on host type
    - Data on users with failed and sucessful logons

    """

    metadata = _CLS_METADATA

    @set_text(docs=_CELL_DOCS, key="run")  # noqa: MC0001, C901
    # pylint: disable=too-many-locals, too-many-branches, too-many-statements
    def run(  # noqa:MC0001, C901
        self,
        value: Any = None,
        data: Optional[pd.DataFrame] = None,
        timespan: Optional[TimeSpan] = None,
        options: Optional[Iterable[str]] = None,
        **kwargs,
    ) -> HostLogonsSummaryResult:
        """
        Return host summary data.

        Parameters
        ----------
        value : str
            Host name
        data : Optional[pd.DataFrame], optional
            Optionally pass raw data to use for analysis, by default None
        timespan : TimeSpan
            Timespan over which operations such as queries will be
            performed, by default None.
            This can be a TimeStamp object or another object that
            has valid `start`, `end`, or `period` attributes.
            Alternatively you can pass `start` and `end` datetime objects.
        options : Optional[Iterable[str]], optional
            List of options to use, by default None
            A value of None means use default options.

        Returns
        -------
        HostLogonsSummaryResults
            Result object with attributes for each result type.

        Raises
        ------
        MsticnbMissingParameterError
            If required parameters are missing

        MsticnbDataProviderError
            If data is not avaliable

        """
        super().run(
            value=value, data=data, timespan=timespan, options=options, **kwargs
        )

        # Check we have at either a dataset or a host_name and timespan
        if not (data is not None or (value is not None and timespan is not None)):
            raise MsticnbMissingParameterError("data, or a hostname and timespan.")

        # If data is not provided use host_name and timespan to get data
        if not isinstance(data, pd.DataFrame) or data.empty and timespan:
            nb_data_wait(f"{value}")
            host_verif = verify_host_name(
                qry_prov=self.query_provider, timespan=timespan, host_name=value
            )
            if host_verif.host_names:
                nb_print(f"Could not obtain unique host name from {value}. Aborting.")
                return self._last_result
            if not host_verif.host_name:
                nb_print(
                    f"Could not find event records for host {value}. "
                    + "Results may be unreliable."
                )
                # return self._last_result
            host_type = host_verif.host_type or None
            host_name = host_verif.host_name or value

            if host_type == "Windows" or not host_type == "Linux":
                # If no known data type try Windows
                data = self.query_provider.WindowsSecurity.list_all_logons_by_host(  # type: ignore
                    host_name=host_name, start=timespan.start, end=timespan.end  # type: ignore
                )
            else:
                data = self.query_provider.LinuxSyslog.list_logons_for_host(  # type: ignore
                    host_name=host_name, start=timespan.start, end=timespan.end  # type: ignore
                )
        else:
            # If data is provided do some required formatting
            data = _format_raw_data(data)

        # Add description to results for context
        self._last_result = HostLogonsSummaryResult(
            notebooklet=self, description=self.metadata.description, timespan=timespan
        )

        # Check we have data
        if not isinstance(data, pd.DataFrame) or data.empty:
            print("No valid data avaliable.")
            return self._last_result

        # Conduct analysis and get visualizations
        nb_print("Performing analytics and generating visualizations")
        logon_sessions_df = data[data["LogonResult"] != "Unknown"]

        if "timeline" in self.options:
            tl_plot = _gen_timeline(data, self.silent)
            self._last_result.timeline = tl_plot

        if "map" in self.options:
            logon_map = _map_logons(data, self.silent)
            self._last_result.logon_map = logon_map
        logon_matrix = _logon_matrix(data, self.silent)
        if "charts" in self.options:
            pie = _users_pie(data, self.silent)
            stack_bar = _process_stack_bar(data, self.silent)
            charts = {"User Pie Chart": pie, "Process Bar Chart": stack_bar}
            self._last_result.plots = charts
        if "failed_success" in self.options:
            failed_success_df = _failed_success_user(data, self.silent)
            self._last_result.failed_success = failed_success_df

        self._last_result.logon_sessions = logon_sessions_df
        self._last_result.logon_matrix = logon_matrix

        return self._last_result
        # pylint: enable=too-many-locals, too-many-branches


@set_text(docs=_CELL_DOCS, key="logons_timeline")
def _gen_timeline(data: pd.DataFrame, silent: bool):
    if silent:
        return timeline.display_timeline(
            data[data["LogonResult"] != "Unknown"],
            group_by="LogonResult",
            source_columns=[
                "Account",
                "LogonProcessName",
                "SourceIP",
                "LogonTypeName",
                "LogonResult",
            ],
            hide=True,
            title="Logon Events Over Time - Grouped by Result.",
        )

    return timeline.display_timeline(
        data[data["LogonResult"] != "Unknown"],
        group_by="LogonResult",
        source_columns=[
            "Account",
            "LogonProcessName",
            "SourceIP",
            "LogonTypeName",
            "LogonResult",
        ],
        title="Logon Events Over Time - Grouped by Result.",
    )


@set_text(docs=_CELL_DOCS, key="show_map")
def _map_logons(data: pd.DataFrame, silent: bool) -> FoliumMap:
    """Produce a map of source IP logon locations."""
    map_data = data[~(data["IpAddress"].isin(["-", "::1", "", "NaN"]))]
    if not isinstance(map_data, pd.DataFrame) or map_data.empty:
        if not silent:
            md("No plottable logins available")
        return None
    if not silent:
        display(
            map_data.mp_plot.folium_map(
                ip_column="IpAddress", layer_column="LogonResult"
            )
        )

    return map_data.mp_plot.folium_map(
        ip_column="IpAddress", layer_column="LogonResult"
    )


@set_text(docs=_CELL_DOCS, key="show_pie")
def _users_pie(data: pd.DataFrame, silent: bool) -> figure:
    """Produce pie chart based on observance of user names in data."""
    output_notebook()
    user_logons = (
        data["Account"]
        .value_counts()
        .head(20)
        .reset_index(name="value")
        .rename(columns={"index": "Account"})
    )
    # Exclude blank user names as Linux data can be skewed by this
    if "" in user_logons:
        user_logons.drop("", inplace=True)
    user_logons["angle"] = user_logons["value"] / user_logons["value"].sum() * 2 * pi
    user_logons["color"] = viridis(len(user_logons))

    viz = figure(
        height=350,
        title="20 most prevelent users",
        toolbar_location=None,
        tools="hover",
        tooltips="@Account: @value",
        x_range=(-0.5, 1.0),
    )

    viz.wedge(
        x=0,
        y=1,
        radius=0.4,
        start_angle=cumsum("angle", include_zero=True),
        end_angle=cumsum("angle"),
        line_color="white",
        fill_color="color",
        legend_field="Account",
        source=user_logons,
    )

    viz.axis.axis_label = None
    viz.axis.visible = False
    viz.grid.grid_line_color = None

    if not silent:
        show(viz)

    return viz


# pylint: disable=too-many-locals
@set_text(docs=_CELL_DOCS, key="show_process")
def _process_stack_bar(data: pd.DataFrame, silent: bool) -> figure:
    """Produce stacked bar chart showing logon result by process."""
    proc_list = data["LogonTypeName"].unique()
    s_data = []
    f_data = []
    procs = []
    for process in proc_list:
        procs.append(process)
        proc_events = data[data["LogonTypeName"] == process][
            "LogonResult"
        ].value_counts()
        try:
            s_count = proc_events["Success"]
        except KeyError:
            s_count = 0
        try:
            f_count = proc_events["Failure"]
        # Catch error caused if one type of logon result is missing
        except KeyError:
            f_count = 0
        # Convert counts to a percentage to get a more useful output
        if (s_count + f_count) > 0:
            s_per = (s_count / (f_count + s_count)) * 100
            f_per = 100 - s_per
            s_data.append(s_per)
            f_data.append(f_per)
        else:
            procs.remove(process)
    processes = procs
    results = ["Success", "Failure"]
    colors = ["#536d4c", "#832828"]

    graph_data = {"processes": procs, "Success": s_data, "Failure": f_data}

    viz = figure(
        x_range=processes,
        height=350,
        title="Logon Result % by Logon Type",
        toolbar_location=None,
        tools="hover",
        tooltips="@processes $name: @$name%",
    )

    viz.vbar_stack(
        results,
        x="processes",
        width=0.75,
        color=colors,
        source=graph_data,
        legend_label=results,
    )

    viz.y_range.start = 0  # type: ignore[attr-defined]
    viz.x_range.range_padding = 0.1  # type: ignore[attr-defined]
    viz.xgrid.grid_line_color = None  # type: ignore[attr-defined]
    viz.axis.minor_tick_line_color = None
    viz.yaxis.axis_label = "% of logons"
    viz.xaxis.axis_label = "Process name"  # type: ignore[assignment]
    viz.outline_line_color = None  # type: ignore[assignment]
    viz.legend.location = "top_left"
    viz.legend.orientation = "horizontal"

    if not silent:
        show(viz)

    return viz


# pylint: enable=too-many-locals


@set_text(docs=_CELL_DOCS, key="logon_matrix")
def _logon_matrix(data: pd.DataFrame, silent: bool) -> pd.DataFrame:
    """Produce DataFrame showing logons grouped by user and process."""
    print(data.columns)
    logon_by_type = (
        data[(data["Account"] != "") & (data["LogonResult"] != "Unknown")][  # type: ignore
            ["Account", "LogonTypeName", "LogonResult", "TimeGenerated"]
        ]
        .groupby(["Account", "LogonTypeName", "LogonResult"])
        .count()
        .unstack()
        .sort_values(by=[("TimeGenerated", "Success")])
        .rename(columns={"EventID": "LogonCount"})
        .fillna(0)
        .style.background_gradient(cmap="viridis", low=0.5, high=0)
        .format("{0:0>3.0f}")
    )
    if not silent:
        display(logon_by_type)
    return logon_by_type


@set_text(docs=_CELL_DOCS, key="show_df")
def _failed_success_user(data: pd.DataFrame, silent: bool) -> pd.DataFrame:
    failed_logons = data[(data["LogonResult"] == "Failure") & (data["Account"] != "")]
    host_logons = data[(data["LogonResult"] == "Success") & (data["Account"] != "")]
    combined = host_logons[
        host_logons["Account"].isin(failed_logons["Account"].drop_duplicates())
    ]

    if not silent:
        if combined.empty:
            nb_print("No accounts with both a successful and failed logon on.")
        else:
            display(combined)
    return combined


def _win_remote_ip(row: pd.Series) -> str:
    """Return remote IP address from Windows log if logon type is remote."""
    return row["IpAddress"] if row["LogonType"] == 3 else "NaN"


def _format_raw_data(data: pd.DataFrame) -> pd.DataFrame:
    """Populate required fields in DataFrame if they aren not included."""
    if "SourceSystem" in data.columns and data["SourceSystem"].iloc[0] == "Linux":
        if "LogonResult" not in data.columns:
            data["LogonResult"] = data.apply(_get_logon_result_lx, axis=1)
            data["Account"] = data.apply(_parse_user_lx, axis=1)
            data["SourceIP"] = data.apply(_parse_ip_lx, axis=1)
        if "LogonProcessName" not in data.columns:
            data.rename(columns={"ProcessName": "LogonProcessName"})
        if "LogonTypeName" not in data.columns:
            data["LogonTypeName"] = data["LogonProcessName"]
    elif "SubjectUserSid" in data.columns:
        if "LogonResult" not in data.columns:
            data["LogonResult"] = data.apply(_event_id_to_result, axis=1)
            data["SourceIP"] = data.apply(_win_remote_ip, axis=1)
    return data


def _get_logon_result_lx(row: pd.Series) -> str:
    """Identify if a Linux syslog event is for a sucessful or failed logon."""
    failure_events = row.str.contains(
        """failure|failed|invalid|unable to negotiate|authentication failures|did not receive identification|bad protocol version identification|^Connection closed .* [preauth]""",  # pylint: disable=line-too-long
        regex=True,
    )

    success_events = row.str.contains("Accepted|Successful", regex=True)
    if failure_events["SyslogMessage"] is True:
        return "Failure"
    return "Success" if success_events["SyslogMessage"] is True else "Unknown"


def _parse_user_lx(row: pd.Series) -> Union[str, Any, None]:
    """Extract a user name from Syslog message related to a logon."""
    if row.str.contains("publickey")["SyslogMessage"] is True:
        regex = re.compile("for ([^ ]*)")
        user = re.search(regex, row["SyslogMessage"])
        return user[1] if user else None

    regex = re.compile("user |user= ([^ ]*)")
    user = re.search(regex, row["SyslogMessage"])
    return user[1] if user else None


def _parse_ip_lx(row: pd.Series) -> Union[str, Any, None]:
    """Extract an IP Address from an Syslog message."""
    regex = re.compile("((?:[0-9]{1,3}\\.){3}[0-9]{1,3})")
    ips = re.search(regex, row["SyslogMessage"])
    return ips[1] if ips else None


def _event_id_to_result(row: pd.Series):
    """Transform Windows event id to string for logon results."""
    if row["EventID"] == 4624:
        return "Success"
    return "Failure" if row["EventID"] == 4625 else "Unknown"



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Functions to create documentation from notebooklets classes."""
import html
import inspect
from typing import List

from markdown import markdown

from ._version import VERSION
from .notebooklet import Notebooklet, NotebookletResult

__version__ = VERSION
__author__ = "Ian Hellen"


def get_class_doc(doc_cls: type, fmt: str = "html") -> str:
    """
    Create HTML documentation for the notebooklet class.

    Parameters
    ----------
    doc_cls : type
        The class to document
    fmt : str
        Format = "html" or "md", by default "html"

    Returns
    -------
    str
        HTML documentation for the class

    Raises
    ------
    TypeError
        If the class is not a subclass of Notebooklet.

    """
    if not issubclass(doc_cls, Notebooklet):
        raise TypeError("doc_cls must be a type of Notebooklet")
    if fmt == "html":
        return markdown(_get_main_class_doc_md(doc_cls))
    return _get_main_class_doc_md(doc_cls)


def _get_main_class_doc_md(doc_cls) -> str:
    """Return Markdown format of class documentation."""
    cls_doc_lines = [f"# Notebooklet Class - {doc_cls.__name__}\n"]
    cls_doc_str = inspect.getdoc(doc_cls)
    if cls_doc_str:
        fmt_doc_lines: List[str] = []
        for idx, doc_line in enumerate(inspect.cleandoc(cls_doc_str).split("\n")):
            if doc_line.strip().startswith("--") and idx > 0:
                # if this is a heading underline, bold the previous line
                fmt_doc_lines[idx - 1] = f"**{fmt_doc_lines[idx - 1]}**"
                fmt_doc_lines.append("")
            else:
                fmt_doc_lines.append(doc_line + "\n")
        cls_doc_lines.extend(fmt_doc_lines)
    cls_doc_lines.append("\n---\n")

    cls_doc_lines.append("## Display Sections")
    for _, func in inspect.getmembers(doc_cls, inspect.isfunction):
        cls_doc_lines.extend(_get_closure_vars(func, doc_cls))

    for _, func in inspect.getmembers(inspect.getmodule(doc_cls), inspect.isfunction):
        cls_doc_lines.extend(_get_closure_vars(func, doc_cls))

    cls_doc_lines.append("\n---\n")
    cls_doc_lines.append("## Results Class\n")
    for cls_name, cls in inspect.getmembers(
        inspect.getmodule(doc_cls), inspect.isclass
    ):
        if issubclass(cls, NotebookletResult) and cls is not NotebookletResult:
            cls_doc_lines.append(f"## {cls_name}\n")
            cls_doc_lines.append(_get_result_doc(cls))
            break
    cls_doc_lines.append("\n---\n")
    cls_doc_lines.append("## Methods")
    cls_doc_lines.append("### Instance Methods")
    cls_doc_lines.append(_get_class_methods_doc(doc_cls))
    cls_doc_lines.append("### Other Methods")
    cls_doc_lines.append(_get_class_func_doc(doc_cls))
    return "\n".join(cls_doc_lines)


def _get_closure_vars(func, doc_cls) -> List[str]:
    """Return title and text from function args."""
    cls_doc_lines = []
    closure_args = inspect.getclosurevars(func).nonlocals

    # If the function is using the metadata docs and key
    # try to fetch that from the class module
    docs = closure_args.get("docs")
    key = closure_args.get("key")
    other_items = None
    title = text = None
    hd_level = 2
    if docs and key and issubclass(doc_cls, Notebooklet):
        cell_docs = getattr(doc_cls, "_cell_docs", None)
        if cell_docs:
            title = cell_docs.get(key, {}).get("title")
            text = cell_docs.get(key, {}).get("text")
            hd_level = cell_docs.get(key, {}).get("hd_level", 2) + 1
            other_items = {
                hdr: str(text)
                for hdr, text in docs.get(key, {}).items()
                if hdr not in ("title", "text", "hd_level", "md")
            }
    else:
        # Otherwise use inline parameters
        title = closure_args.get("title")
        hd_level = closure_args.get("hd_level", 2) + 1
        text = closure_args.get("text")
    if title:
        cls_doc_lines.append(("#" * hd_level) + f" {title}\n")
        if text:
            cls_doc_lines.append(text)
    if other_items:
        for name, content in other_items.items():
            cls_doc_lines.append(f"**{name}**\n{content}")
    return cls_doc_lines


def _get_result_doc(cls) -> str:
    """Return Markdown documentation for Result class."""
    attr_section = False
    doc_lines = []
    cls_doc_str = inspect.getdoc(cls)
    if not cls_doc_str:
        return ""
    for line in inspect.cleandoc(cls_doc_str).split("\n"):
        if line.startswith("---"):
            attr_section = True

        elif attr_section:
            line = "\n- " + line + "<br>" if not line.startswith(" ") else line.strip()
        doc_lines.append(line)
    return "\n".join(doc_lines)


def _get_class_methods_doc(doc_cls: type) -> str:
    """Get class instance methods."""
    doc_lines: List[str] = []
    doc_lines_parent: List[str] = []
    allow_inherited = ["__init__", "run"]
    nb_methods = [
        f_name for f_name, _ in inspect.getmembers(Notebooklet, inspect.isfunction)
    ]
    cls_methods = inspect.getmembers(doc_cls, inspect.isfunction)

    for func_name, func in sorted(cls_methods):
        # First list is:
        # - run and __init__ methods
        # - other subclass funcs that are not in the parent class
        # (but not if it's a private method)
        if func_name in allow_inherited or (
            func_name not in nb_methods and not func_name.startswith("_")
        ):
            doc_lines.extend(_format_func_doc(func_name, func, True))
        elif func_name in nb_methods and not func_name.startswith("_"):
            doc_lines_parent.extend(_format_func_doc(func_name, func, True))
    doc_lines.append("## Inherited methods")
    doc_lines.extend(doc_lines_parent)
    return "\n".join(doc_lines)


def _get_class_func_doc(doc_cls: type) -> str:
    """Get class functions (class methods and properties)."""
    doc_lines: List[str] = []
    prop_set = {
        f_name
        for f_name, _ in inspect.getmembers(doc_cls, lambda f: isinstance(f, property))
    }

    def member_crit(member):
        return inspect.ismethod(member) or isinstance(member, property)

    for func_name, func in inspect.getmembers(doc_cls, member_crit):
        if not func_name.startswith("_"):
            doc_lines.extend(_format_func_doc(func_name, func, False, prop_set))
    return "\n".join(doc_lines)


def _format_func_doc(func_name, func, full_doc=False, prop_set=None):
    """Format function signature."""
    func_disp_name = func_name.replace("_", "\\_")
    doc_lines = [f"#### {func_disp_name}\n"]
    if prop_set and func_name in prop_set:
        doc_lines.append(f"{func_disp_name} [property]")
    else:
        func_sig = html.escape(str(inspect.signature(func)))
        doc_lines.append(f"{func_disp_name}{func_sig}<br>")

    func_doc = inspect.getdoc(func)
    if func_doc:
        if not full_doc:
            # Get the first line of the doc string
            doc_lines.append(func_doc.split("\n", maxsplit=1)[0])
            return doc_lines

        func_doc = inspect.cleandoc(func_doc).split("\n", maxsplit=1)[0]
    if func_doc:
        refmt_headings = []
        for doc_line in func_doc.split("\n"):
            if doc_line.startswith("### "):
                refmt_headings.append(doc_line.replace("### ", "##### "))
            elif doc_line.startswith("## "):
                refmt_headings.append(doc_line.replace("## ", "#### "))
            else:
                refmt_headings.append(doc_line + "\n")
        doc_lines.extend(refmt_headings)
    return doc_lines



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Setup script for msticnb."""

import re
import setuptools


with open("requirements.txt", "r", encoding="utf-8") as fh:
    INSTALL_REQUIRES = fh.readlines()

# pylint: disable=locally-disabled, invalid-name
with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

with open("msticnb/_version.py", "r", encoding="utf-8") as fd:
    v_match = re.search(r'^VERSION\s*=\s*[\'"]([^\'"]*)[\'"]', fd.read(), re.MULTILINE)
    __version__ = v_match.group(1) if v_match else "no version"
# pylint: enable=locally-disabled, invalid-name

setuptools.setup(
    name="msticnb",
    version=__version__,
    author="Ian Hellen",
    author_email="ianhelle@microsoft.com",
    description="MSTIC Notebooklets",
    license="MIT License",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/microsoft/msticnb",
    project_urls={
        "Documentation": "https://msticnb.readthedocs.io",
        "Code": "https://github.com/microsoft/msticnb",
    },
    python_requires=">=3.6",
    packages=setuptools.find_packages(exclude=["tests.*", "tests"]),
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 4 - Beta",
    ],
    install_requires=INSTALL_REQUIRES,
    keywords=[
        "security",
        "cybersecurity",
        "infosec",
        "jupyter",
        "notebook",
        "azure",
        "sentinel",
    ],
    zip_safe=False,
    include_package_data=True,
)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""IP Helper functions."""

import contextlib
from typing import Any, Dict, List, Optional, Tuple, Union

import pandas as pd
from msticpy.datamodel.entities import IpAddress

try:
    from msticpy.context.ip_utils import get_whois_df
    from msticpy.vis.foliummap import FoliumMap
except ImportError:
    # Fall back to msticpy locations prior to v2.0.0
    from msticpy.nbtools.foliummap import FoliumMap
    from msticpy.sectools.ip_utils import get_whois_df

from .._version import VERSION
from ..common import nb_markdown
from .ti import get_ti_results

__version__ = VERSION
__author__ = "Ian Hellen"


def get_ip_ti(
    ti_lookup,
    data: pd.DataFrame,
    ip_col: str,
) -> Tuple[Any, Optional[pd.DataFrame]]:
    """
    Lookup Threat Intel for IPAddress.

    Parameters
    ----------
    ti_lookup : TILookup
        TI Lookup provider
    data : pd.DataFrame
        Input data frame
    ip_col : str
        Name of Ip address column

    Returns
    -------
    pd.DataFrame
        DataFrame with TI results for IPs

    """
    return get_ti_results(ti_lookup, data, ip_col)


def _normalize_ip4(data, ip_col):
    ip4_rgx = r"((?:[0-9]{1,3}\.){3}[0-9]{1,3})"
    ip4_match = (data[ip_col].notna()) & (data[ip_col].str.match(ip4_rgx))
    ipv4_df = data[ip4_match]
    other_df = data[~ip4_match]
    ipv4_df = (
        ipv4_df.assign(IP_ext=lambda x: x[ip_col].str.extract(ip4_rgx, expand=False))
        .rename(columns={ip_col: ip_col + "_orig"})
        .rename(columns={"IP_ext": ip_col})
    )
    return pd.concat([ipv4_df, other_df])


def get_geoip_whois(geo_lookup, data: pd.DataFrame, ip_col: str):
    """
    Get GeoIP and WhoIs data for IPs.

    Parameters
    ----------
    geo_lookup : GeoIpLookup
        GeoIP Provider
    data : pd.DataFrame
        Input data frame
    ip_col : str
        Name of Ip address column

    Returns
    -------
    pd.DataFrame
        Results dataframe with GeoIP and WhoIs data

    """
    data = _normalize_ip4(data, ip_col)
    nb_markdown(f"Querying geolocation for {len(data)} ip addresses...")

    geo_ips = geo_lookup.lookup_ips(data, column=ip_col)
    geo_df = data.merge(geo_ips, how="left", left_on=ip_col, right_on="IpAddress")

    nb_markdown(f"Querying WhoIs for {len(data)} ip addresses...")
    # Get the WhoIs results
    return get_whois_df(geo_df, "IpAddress", whois_col="Whois_data")


def map_ips(
    data: pd.DataFrame,
    ip_col: str,
    summary_cols: Optional[List[str]] = None,
    geo_lookup: Any = None,
) -> FoliumMap:
    """
    Produce a map of IP locations.

    Parameters
    ----------
    geo_lookup: Any
        Geo-IP provider instance
    data : pd.DataFrame
        Data containing the IPAddress
    ip_col : str
        [description]
    summary_cols : Optional[List[str]], optional
        [description], by default None
    geo_lookup : Any
        GeoIP Provider instance.

    Returns
    -------
    FoliumMap
        Folium map with items plotted.

    """
    data = _normalize_ip4(data, ip_col=ip_col)
    ip_ent_list = []
    for _, row in data.iterrows():
        if not isinstance(row[ip_col], str) or row[ip_col] in ("", "NaN", "-"):
            continue
        ip_ent_rslt = convert_to_ip_entities(row[ip_col], geo_lookup=geo_lookup)
        if summary_cols:
            for ip_ent in ip_ent_rslt:
                for col in summary_cols:
                    ip_ent.AdditionalData[col] = row[col]
        ip_ent_list.extend(ip_ent_rslt)

    # Get center point of logons and build map account that
    folium_map = FoliumMap(zoom_start=4)
    folium_map.center_map()
    if ip_ent_list:
        icon_props = {"color": "green"}
        folium_map.add_ip_cluster(ip_entities=ip_ent_list, **icon_props)
    return folium_map


def convert_to_ip_entities(  # noqa: MC0001
    ip_str: Optional[str] = None,
    data: Optional[pd.DataFrame] = None,
    ip_col: Optional[str] = None,
    geo_lookup: Any = None,
) -> List[IpAddress]:
    """
    Take in an IP Address string and converts it to an IP Entity.

    Parameters
    ----------
    ip_str : str
        A string with a single IP Address or multiple addresses
        delimited by comma or space
    data : pd.DataFrame
        Use DataFrame as input
    ip_col : str
        Column containing IP addresses
    geo_lookup : bool
        If true, do geolocation lookup on IPs,
        by default, True

    Returns
    -------
    List
        The populated IP entities including address and geo-location

    Raises
    ------
    ValueError
        If neither ip_string or data/column provided as input

    """
    ip_entities: List[IpAddress] = []
    all_ips: Dict[str, IpAddress] = {}

    if ip_str:
        addrs = arg_to_list(ip_str)
    elif data is not None and ip_col:
        addrs = list(data[ip_col].values)
    else:
        raise ValueError("Must specify either ip_str or data + ip_col parameters.")

    for addr in addrs:
        if isinstance(addr, list):
            ip_list = set(addr)
        elif isinstance(addr, str) and "," in addr:
            ip_list = {ip.strip() for ip in addr.split(",")}
        else:
            ip_list = {addr}

        for ip_addr in ip_list:
            if ip_addr in all_ips:
                ip_entities.append(all_ips[ip_addr])
                continue
            ip_entity = IpAddress(Address=ip_addr)

            if geo_lookup:
                # pylint: disable=broad-except
                # lookup_ip might raise a number of exception types
                # we - just want to ignore here
                with contextlib.suppress(Exception):
                    geo_lookup.lookup_ip(ip_entity=ip_entity)
            ip_entities.append(ip_entity)
    return ip_entities


def arg_to_list(arg: Union[str, List[str]], delims=",; ") -> List[str]:
    """
    Convert an optional list/str/str with delims into a list.

    Parameters
    ----------
    arg : Union[str, List[str]]
        A string, delimited string or list
    delims : str, optional
        The default delimiters to use, by default ",; "

    Returns
    -------
    List[str]
        List of string components

    Raises
    ------
    TypeError
        If `arg` is not a string or list

    """
    if isinstance(arg, list):
        return arg
    if isinstance(arg, str):
        for char in delims:
            if char in arg:
                return [item.strip() for item in arg.split(char)]
        return [arg]
    raise TypeError("`arg` must be a string or a list.")



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Data viewers mixin classes."""
from typing import List, Optional

import numpy as np
import pandas as pd

try:
    from msticpy.nbwidgets import SelectItem
except ImportError:
    # Fall back to msticpy locations prior to v2.0.0
    from msticpy.nbtools.nbwidgets import SelectItem

from ._version import VERSION
from .common import MsticnbMissingParameterError

__version__ = VERSION
__author__ = "Ian Hellen"


class DFViewer:
    """Mixin class for NotebookletResult."""

    def view_events(
        self,
        summary_cols: Optional[List[str]] = None,
        attrib: Optional[str] = None,
        data: Optional[pd.DataFrame] = None,
        **kwargs,
    ) -> SelectItem:
        """
        Return simple data view for DataFrame/result attribute.

        Parameters
        ----------
        summary_cols : List[str], optional
            [description]
        attrib : Optional[str], optional
            [description], by default None
        data : Optional[pd.DataFrame], optional
            [description], by default None
        kwargs :
            Additional keyword arguments passed to the SelectItem
            widget.

        Returns
        -------
        SelectItem
            Browser for events in DataFrame.

        Raises
        ------
        AttributeError
            Attribute name not in results class.
        TypeError
            Input data or attribute is not a DataFrame
        MsticnbMissingParameterError
            One of `data` or `attrib` parameters must be supplied
        KeyError
            Summary column name specified that isn't in the DataFrame

        """
        if attrib:
            data = getattr(self, attrib, None)
            if data is None:
                raise AttributeError(f"Attribute '{attrib}' not found in this result.")
        if data is None:
            raise MsticnbMissingParameterError("'attrib' or 'data'")
        if not isinstance(data, pd.DataFrame):
            raise TypeError("The 'data' or 'attrib' must be a DataFrame.")
        summary_cols = summary_cols or list(data.columns)[:3]

        missing_cols = [col for col in summary_cols if col not in data.columns]
        if missing_cols:
            raise KeyError(
                f"Column(s) not found in the data: {', '.join(missing_cols)}"
            )
        if "TimeGenerated" in data.columns and "TimeGenerated" not in summary_cols:
            summary_cols = ["TimeGenerated", *summary_cols]

        pd.set_option("display.max_rows", 150)
        sel_item_params = {
            "item_dict": self._create_options(data, columns=summary_cols),
            "action": self._create_view_callback(data),
            "width": "70%",
            "height": "300px",
        }
        sel_item_params.update(kwargs)
        return SelectItem(**sel_item_params)

    @staticmethod
    def _create_view_callback(data):
        def _event_display(index):
            return pd.DataFrame(
                data.loc[index].replace(r"^\s*$", np.nan, regex=True).dropna().T
            )

        return _event_display

    @staticmethod
    def _create_options(data, columns):
        data_dict = data[columns].to_dict(orient="index")
        return {
            " - ".join(str(item) for item in data.values()): idx
            for idx, data in data_dict.items()
        }
